/*-------------------------------------------------
 *    Kernel version 2.4.20 - 2.6.8, 2.6.13
 *    RS485 (COM-port)
 * 
 *    Andre Zaitsev, Olga Voziakova
 *    ovoz@sai.msu.ru
 *    $History:
 *
 *---------------------------------------- */
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kdev_t.h>
#include <linux/fs.h>

#include <linux/errno.h>
#include <linux/slab.h>
#include <asm/uaccess.h>    //  copy_from_user
#include <linux/ioport.h>
#include <linux/delay.h>    // udelay
#include <linux/interrupt.h>
#include <asm/irq.h>
#include <asm/io.h>

#include <mcua.h>         // To define KERN_2_6
#ifdef KERN_2_6
#include <linux/cdev.h>
#endif

#include <com.h>
#include <rs485.h>
#include <rs485_tools.h>
#include <rs485_irq.h> 
#include <rserrno.h>

static unsigned int major;
#ifdef KERN_2_6
static dev_t dev;
static struct cdev *cdev;
static unsigned int rs_nr_dev = 2;
static unsigned int minor = 0;
#endif

static int io1 = 0;
static int io2 = 0;

int irq1 = -1;
int irq2 = -1;

static int baudrate1 = 0;
static int baudrate2 = 0;

struct file_operations rs485_fops;
RS485INFO *info_port1, *info_port2;

rs485_time_t end_packet_out_tm;
rs485_time_t start_packet_inp_tm;
rs485_time_t tm;

inline int rs485_getIO( int port) {
    return ( port==2 )? _IO_COM2 : _IO_COM1;
}

inline int rs485_getIRQ( int port ) {
    return ( port==2 )? _IRQ_COM2 : _IRQ_COM1;
}  

//inline int rs485_get_port(int irq) {
//    return ( irq == irq1 )? 1 : 2;
//}


inline void rs485_allow_interrupt_card(RS485INFO *pinfo) {
    outb( UART_IER_RDI, pinfo->base +UART_IER );
}

inline void rs485_stop_interrupt_card(RS485INFO *pinfo) {
    outb( 0, pinfo->base +UART_IER );
}

inline void rs485_transmit_disable(RS485INFO *pinfo) {
    outb( RXENABLE, pinfo->base + RX_TX_REG );
}

inline void rs485_transmit_enable(RS485INFO *pinfo) {
    outb( TXENABLE, pinfo->base + RX_TX_REG );
}

int rs485_test_interface(RS485INFO *pinfo) {
    rs485_transmit_disable(pinfo); // ?? 
    if( ( inb( pinfo->base + UART_MSR ) & UART_MSR_CTS ) == 0 ) return -ERSNOINTERFACE; 
    return SUCCESS;
}

int rs485_test_power( RS485INFO *pinfo ) {
    rs485_stop_interrupt_card( pinfo );
    rs485_transmit_enable( pinfo );
    inb_p(pinfo->base);
    outb_p( 0x55, pinfo->base );  // Test byte
    rs485_set_starttime();
    while( ( inb_p( pinfo->base + UART_LSR ) & UART_LSR_DR ) == 0 ) {
        if( rs485_timeout( TBYTE_TIMEOUT ) ) {
            rs485_transmit_disable(pinfo);
            rs485_allow_interrupt_card( pinfo );
	    PDEBUG( "test_power: timeout: no echo!\n" );
	    return -ERSSHUTDOWN;   // no echo => no power
        }
        udelay( TBYTE_WAIT );
    }
    if( inb_p(pinfo->base) != 0x55 ) PDEBUG( "rs485_test_power: BAD ECHO\n");
    rs485_transmit_disable( pinfo );
    rs485_allow_interrupt_card( pinfo );
    return SUCCESS;
}

int rs485_test_device(RS485INFO *pinfo) {
    int ret = SUCCESS;
    if( ( ret = rs485_test_interface(pinfo) ) != SUCCESS ) return ret;
    return rs485_test_power( pinfo );
}

static unsigned char uart_ier_back = 0;
static unsigned char uart_lcr_back = 0;
static unsigned char uart_mcr_back = 0;

void rs485_get_reg(RS485INFO *pinfo) {
    uart_ier_back = inb_p(pinfo->base +UART_IER);
    uart_lcr_back = inb_p(pinfo->base +UART_LCR);
    uart_mcr_back = inb_p(pinfo->base +UART_MCR);
}

void rs485_restore_reg(RS485INFO *pinfo) {
    outb_p(uart_ier_back, pinfo->base +UART_IER);
    outb_p(uart_lcr_back, pinfo->base +UART_LCR);
    outb_p(uart_mcr_back, pinfo->base +UART_MCR);
}

int ubrr_to_rate(int baudrate_code) {
    if( baudrate_code == 0 ) return 0;
    return(Fosc/16/baudrate_code);
}

int rate_to_ubrr(int baudrate) {
    int baudrate_code;
    baudrate = baudrate < Minbaudrate ? Minbaudrate : baudrate;
    baudrate = baudrate > Maxbaudrate ? Maxbaudrate : baudrate;
    baudrate_code = Fosc/16/baudrate;
    return baudrate_code;
}

int rs485_set_baudrate(RS485INFO *pinfo) {
    int baudrate_code;
    baudrate_code = rate_to_ubrr( pinfo->baudrate );
    pinfo->baudrate = ubrr_to_rate( baudrate_code );
    PDEBUG( "set baudrate=%d\n", pinfo->baudrate );
    outb_p( 0x80, pinfo->base + UART_LCR );
    outw( baudrate_code, pinfo->base +UART_DLL ); 
    outb_p( OTHER, pinfo->base + UART_LCR );
    return SUCCESS;
}

int rs485_reset(RS485INFO *pinfo) {    //  -
    rs485_transmit_disable(pinfo);
    msleep( 8 );
    outb_p( UART_FCR_ENABLE_FIFO|UART_FCR_CLEAR_RCVR|UART_FCR_CLEAR_XMIT|UART_FCR_TRIGGER_1, pinfo->base+UART_FCR );
    inb_p( pinfo->base+UART_RX );
    rs485_set_baudrate( pinfo );
//    rs485_allow_interrupt_card( pinfo );
//    rs485_fifo_flush(pinfo);
    return rs485_test_device( pinfo );
}

// -----------------------------------------------------------------------

inline int getbyte( RS485INFO *info, unsigned char *b, int resp_err ) {
    unsigned short iir;
    *b = 0;
    rs485_set_starttime();
    while( (( iir = inb( info->base+UART_LSR ) ) & UART_LSR_DR ) == 0 ) {
        if( rs485_timeout( TBYTE_TIMEOUT ) ) return resp_err;
        udelay( TBYTE_WAIT );
    }
    *b = inb_p( info->base );        
    if( ( iir & UART_LSR_FE ) != 0 ) {
        PDEBUG( "transmit: frame error!\n" );
        return SignalEER;
    }
    return 0; // OK
}

static unsigned char echo=0;
static int ret = 0;
static volatile unsigned char signal_from_device = 0;

#ifdef __DEBUG_TIMING__
    static rs485_time_t tm_start;
#endif

// debug
static unsigned char last_address=0;

inline void cleanFIFO(RS485INFO *info) {
    outb_p( UART_FCR_ENABLE_FIFO|UART_FCR_CLEAR_RCVR|UART_FCR_CLEAR_XMIT|UART_FCR_TRIGGER_1, info->base+UART_FCR );
    inb( info->base );
}

int start_transmitter( RS485INFO *info ) {
    unsigned int i;
    cleanFIFO(info);
    rs485_stop_interrupt_card(info);
    rs485_stop_interrupts();
    info->rs->status = PASSIVE_DATA;  // it means that PASSIVE_DATA will be wait

// TRANSMITTER:
    outb_p(FIRST, info->base + UART_LCR);
    rs485_transmit_enable(info);
    
    last_address = info->rs->outbuffer.cache[0];
    for(i = 0; i <= info->rs->outbuffer.size; i++) {
        outb_p( info->rs->outbuffer.cache[i], info->base);
        ret = getbyte(info, &echo, -ERSSHUTDOWN);
        if( i == 0 ) outb_p( OTHER, info->base + UART_LCR ); // address-byte
        if( ret == 0 && echo != info->rs->outbuffer.cache[i] ) ret = SignalCOL;
        if( ret != 0 ) {
            PDEBUG("addr=%hx transmit: getbyte timeout, i=%d, info->rs->outbuffer.size=%zd\n",
                                last_address, i, info->rs->outbuffer.size);
            rs485_transmit_disable(info);
            rs485_allow_interrupts();
            rs485_allow_interrupt_card(info);
            return ret;
        }
    }
    rs485_transmit_disable(info);
    rs485_allow_interrupts();
    rs485_allow_interrupt_card(info);  
// REPLY:    
#ifdef __DEBUG_TIMING__
    rs485_gettime(&tm_start);
#endif
    rs485_gettime(&end_packet_out_tm);
    start_packet_inp_tm = end_packet_out_tm;
    tm = end_packet_out_tm;
    rs485_set_starttime();
    
    while( ( info->rs->status & ( SIGNAL_RECEIVED | DATA_READY ) ) == 0 ) {
        if( rs485_timeout( RBYTE_TIMEOUT ) ) {
            if( info->rs->status & ( SIGNAL_RECEIVED | DATA_READY ) ) break;	
            rs485_gettime(&tm);
            rs485_printtm_interval(&end_packet_out_tm, &tm);
            PDEBUG( "transmit: timeout! addr=%hx status=%hx info->rs->inbuffer.size=%zd info->rs->usCountByteInp=%zd\n",
                last_address, info->rs->status, info->rs->inbuffer_pas.size, info->rs->usCountByteInp );
            if( info->rs->status & RECEIVING ) ret = SignalEER;
            else ret = SignalDSC;
            return ret;
        }
        udelay( RBYTE_WAIT );
    }
//

#ifdef __DEBUG_TIMING__
    rs485_time_t tm_end;
    rs485_gettime(&tm_end);
    rs485_printtm_interval(&tm_start, &tm_end);
#endif
//
    if( info->rs->status & SIGNAL_RECEIVED ) {
        info->rs->status &= ~SIGNAL_RECEIVED;
        return signal_from_device;
    }
    if( info->rs->status & DATA_READY ) {
        info->rs->status &= ~DATA_READY;
        if( crc( info->rs->inbuffer_pas.cache, info->rs->inbuffer_pas.size+1 ) == 0 ) {
            ret = info->rs->inbuffer_pas.size;    // size of data without CRC
        } else {
            PDEBUG("addr=%hx Received packet with bad CRC!\n", last_address);
            ret = SignalEER;
        }
        return ret;
    }
    return -(ERSBASE+16);                 // algorithm. error
}                                    

static unsigned int counter = 0;

void receiv_data(RS485INFO * info) {
    unsigned short iir = 0;
    unsigned short csr = 0;
    counter = 0;
    while( ( ( csr = inb( info->base + UART_LSR ) ) & UART_LSR_DR ) != 0 && counter < 32 ) {
        counter++;

// debug
//        if( rs485_timeout(500) ) PDEBUG("counter=%d\n", counter);

        rs485_set_starttime();
        iir = inb( info->base );
        if( info->rs->status & SIGNAL_SENT ) { // Sent signal echo, check it !
            rs485_transmit_disable(info);
            info->rs->status |= DATA_READY;
            info->rs->status &= ~(PASSIVE_DATA|SIGNAL_SENT|RECEIVING);

            if(iir != SignalACK) PDEBUG("Bad echo! %hx != %hx\n", iir, SignalACK);
            if( (inb(info->base + UART_LSR) & UART_LSR_DR ) != 0 ) {
                PDEBUG("receive: addr=%hx echo received, fifo is not empty\n",
                          last_address);
            }
            return;
        }
        if( csr & UART_LSR_FE ) {
            PDEBUG( "receive: frame error, addr=%hx, status=%hx iir=%hx info->rs->usCountByteInp=%zd info->rs->inbuffer.size=%zd\n",
                last_address, info->rs->status, iir, info->rs->usCountByteInp, info->rs->inbuffer_pas.size );

            if( (inb(info->base + UART_LSR) & UART_LSR_DR ) != 0 ) {
                PDEBUG("addr=%hx fifo is not empty\n", last_address);
                cleanFIFO(info);
/*
                for(i=0; i<16; i++) {
                    if( (inb(info->base +UART_LSR) & UART_LSR_DR) != 0 ) {
                        PDEBUG("%hx\n",inb(info->base));
                    } else break;
                }
*/
            }
            return;
        }                            
        if( csr & UART_LSR_PE ) {
            info->rs->usCountByteInp = 0;
            info->rs->status &= ~RECEIVING;
        }
        if( (info->rs->status & RECEIVING) == 0  ) {  // First byte of the packet and PASSIVE mode !!!
            rs485_gettime(&start_packet_inp_tm);
            info->rs->status |= RECEIVING;
            info->rs->usCountByteInp = 0;
            if( iir >= (unsigned char)CACHE_MAXSIZE  ) { // Signal
                info->rs->status |= SIGNAL_RECEIVED;
                signal_from_device = iir;
                info->rs->status &= ~(PASSIVE_DATA|RECEIVING);

                if( (inb(info->base + UART_LSR) & UART_LSR_DR ) != 0 ) {
                    PDEBUG("receive: addr=%hx signal received, fifo is not empty\n",
                       last_address);
                    cleanFIFO(info);
                }
                return;
            }
            if( iir < PACKET_MAX_LENGTH ) {
                info->rs->inbuffer_pas.size = iir + 1;
                info->rs->inbuffer_pas.cache[0] = iir;
                info->rs->usCountByteInp = 1;
                continue;
            } else {
                PDEBUG( "receive: addr=%hx bad length, status=%hx\n", last_address, info->rs->status );
                continue;
            }
        }
        info->rs->inbuffer_pas.cache[info->rs->usCountByteInp] = iir;
        if( info->rs->usCountByteInp+1 > info->rs->inbuffer_pas.size ) {
// SEND ACK:
            outb_p( FIRST, info->base +UART_LCR );
            rs485_transmit_enable( info );
            info->rs->status |= SIGNAL_SENT;
            outb_p( SignalACK, info->base );

            if( (inb(info->base + UART_LSR) & UART_LSR_DR ) != 0 ) {
                PDEBUG("receive: addr=%hx send ACK, fifo is not empty\n", last_address);
                cleanFIFO(info);
            }
            return;
        }
        info->rs->usCountByteInp++;
    } // while-cycle
    return;
}

void rs485_release_resource(void) {
    if( io2 != 0 ) release_region(io2, ADDRESS_REGION_SIZE);
    if( io1 != 0 ) release_region(io1, ADDRESS_REGION_SIZE);
    if( irq2 > 0 ) rs485_clean_irq(info_port2);
    if( (irq1 > 0) && (irq1 != irq2) ) rs485_clean_irq(info_port1);
    return;
}

int rs485_init(void) {        // IO ports and iterrupts initialization
    int err = 0;
#ifndef KERN_2_6
    if( (err=check_region(io1, ADDRESS_REGION_SIZE)) != 0 ) {
        RS_PRINT( KERN_CRIT "%s: I/O port 0x%08lX in use.\n",
                    ADAPTER_ID, (unsigned long) io1);
        io1 = 0;
        irq1 = -1;
        printk("io1=%x failed\n",io1);
    } else {
        request_region(io1, ADDRESS_REGION_SIZE, ADAPTER_ID);
    }
    
    if( (err=check_region(io2, ADDRESS_REGION_SIZE)) != 0 ) {
        RS_PRINT( KERN_CRIT "%s: I/O port 0x%08lX in use.\n",
                    ADAPTER_ID, (unsigned long) io2);
        io2 = 0;
        irq2 = -1;
        printk("io2=%x failed\n",io2);
    } else {
        request_region( io2, ADDRESS_REGION_SIZE, ADAPTER_ID );
    }    
#else  // KERNEL 2.6
    if(!request_region(io1, ADDRESS_REGION_SIZE, ADAPTER_ID)) {
        RS_PRINT( KERN_CRIT "%s: I/O port 0x%08lX in use.\n",
                    ADAPTER_ID, (unsigned long) io1);
        io1 = 0;
        irq1 = -1;
        err = -EACCES;
    }
    if(!request_region(io2, ADDRESS_REGION_SIZE, ADAPTER_ID)) {
        RS_PRINT( KERN_CRIT "%s: I/O port 0x%08lX in use.\n",
                    ADAPTER_ID, (unsigned long) io2);
        io2 = 0;
        irq2 = -1;
        err = -EACCES;
    }    
#endif
    if( (io1 == 0) && (io2 == 0) ) return(err);

    // Interrupt init

    if(irq1 > 0) if( (err=rs485_request_irq(info_port1)) != 0 ) {
        if( io1 != 0 ) release_region(io1, ADDRESS_REGION_SIZE);
        io1 = 0;
        irq1 = -1;
    }
    if((irq2 > 0) && (irq1 != irq2)) {
        if( (err=rs485_request_irq(info_port2)) != 0 ) {
            if( io2 != 0 ) release_region(io2, ADDRESS_REGION_SIZE);
            io2 = 0;
            irq2 = -1;
        }
    }
    if( (irq1 < 0) && (irq2 < 0) ) return(err);
#ifndef KERN_2_6
    if( ( major = register_chrdev( 0, ADAPTER_ID, &rs485_fops ) ) <= 0 ) {
        RS_PRINT( KERN_CRIT "%s: unable to get major\n", ADAPTER_ID);
        rs485_release_resource();
        return major;    //err=major
    } // else RS_PRINT( KERN_INFO "major=%d\n", major );

#else
    if( ( err = alloc_chrdev_region( &dev, minor, rs_nr_dev, ADAPTER_ID ) ) < 0 ) {
        RS_PRINT( KERN_CRIT "%s: unable to get major\n", ADAPTER_ID );
        rs485_release_resource();
        return err;
    }
    major = MAJOR( dev );
//    RS_PRINT( KERN_INFO "major=%d\n", major );

    cdev = cdev_alloc();
    cdev->ops = &rs485_fops;
    cdev->owner = THIS_MODULE;
    err = cdev_add( cdev, dev, rs_nr_dev );
    if( err ) {
        RS_PRINT( KERN_WARNING "rs485: device can't added to the system\n" );
        return err;
    }
#endif
    return SUCCESS;
}

int rs485_open(struct inode *inode, struct file *pfile) {
    RS485INFO *info;
    int ret=0;
//    int imaj, imin;
    
#ifndef KERN_2_6
    switch( MINOR(inode->i_rdev) ) {
#else
    switch( iminor(inode) ) {
#endif
        case 0:
            if( io1 == 0 ) return -ERSBUSY;
            info = info_port1;
            break;
        case 1:
            if( io2 == 0) return -ERSBUSY;
            info = info_port2;
            break;
        default: return -ERSNOPORT;
    }
    
    pfile->private_data = info;

#ifndef KERN_2_6
    PDEBUG("%s: open device %d, %d\n", ADAPTER_ID, MAJOR( inode->i_rdev ), MINOR( inode->i_rdev ) );
//    imaj = MAJOR( inode->i_rdev );
//    imin = MINOR( inode->i_rdev );
#else
    PDEBUG("%s: open device %d, %d\n", ADAPTER_ID, imajor( inode ), iminor( inode ) );
//    imaj = imajor( inode );
//    imin = iminor( inode );
#endif
//    PDEBUG("%s: open device %d, %d\n", ADAPTER_ID, imaj, imin );

    if( info->use_mcua >= MAX_USE_STREAM ) { // More than maximal shared users
        return -ERSBUSY;
    }            
    if( info->use_mcua == 0 ) {    // First opening
        rs485_get_reg( info );
        ret = rs485_reset( info ); 
        if( ret < 0 ) return ret;
    }
    info->use_mcua++;
#ifndef KERN_2_6
    MOD_INC_USE_COUNT;
#endif
    return SUCCESS;
}

int rs485_close(struct inode *inode, struct file *pfile) {

    RS485INFO *info = pfile->private_data;

//    if( down_interruptible(&info->sem ) ) {
//        up(&info->sem);
//        return -EINTR;
//    }

#ifndef KERN_2_6
    MOD_DEC_USE_COUNT;
#endif
    info->use_mcua--;

    if( info->use_mcua == 0 ) {
        rs485_stop_interrupt_card(info);        
        rs485_restore_reg(info);
    }
//    udelay(SEM_WAIT);
//    up(&info->sem);
    return SUCCESS;
}

ssize_t rs485_write( struct file *fp, const char __user *buf, size_t size, loff_t *offset ) {

/*      ( NE   write )
 *    < 0  -  
 *  > 0x80      
 *  < 0x20        , (PACKET_MAX_LENGTH)
 */    
    RS485INFO    *info = fp->private_data;
    unsigned long rest = 0;
    int i, ret = 0;

    if( down_interruptible( &info->sem ) ) {
        up(&info->sem);
        return -EINTR;
    }
    if( (size == 0) || (size > PACKET_MAX_LENGTH ) ) {
        PDEBUG( "rs485_write: size=%zd is invalid!\n", size);
        ret = -ERSFORMAT;
        goto OUT;
    }
    if( (ret=rs485_test_interface(info)) < 0 ) goto OUT;
    rest = copy_from_user( info->rs->outbuffer.cache, buf, size );
    if(rest != 0) {
        PDEBUG("rs485_write: copy_from_user rest=%ld\n", rest);
        ret = -EFAULT;
        goto OUT;
    }
    info->rs->outbuffer.size = size;
    crc( info->rs->outbuffer.cache, size );
    for(i=0; i<MAX_HOP; i++) {
        ret = start_transmitter( info );
        if( (ret != SignalEER) && (ret != SignalNAK) ) break;
        msleep(4);
    }
    if( ( ret > 0 ) && ( ret <= PACKET_MAX_LENGTH ) ) { //   ,    !! 
        rest = copy_to_user( (char*)buf, info->rs->inbuffer_pas.cache, ret);
        if( rest != 0 ) {
            PDEBUG( "rs485_write: copy_to_user rest=%ld\n", rest );
            ret = -EFAULT;
            goto OUT;
        }       
    }
OUT:
    udelay( SEM_WAIT );
    up( &info->sem );
    return ret;
}

ssize_t rs485_read( struct file *fp, char __user *buf,
            size_t count, loff_t *offset ) {
/* 
 * Read data fifo
 */    
  return 0;
}

int rs485_ioctl(struct inode *inode, struct file *fp,
            unsigned int cmd, unsigned long arg) {

    int err = SUCCESS;
    RS485INFO *rs485;
    unsigned long rest = 0;
    
    if(_IOC_TYPE(cmd) != RS_IOC_MAGIC) {        // 2.4 ???
        PDEBUG("ioctl_error!\n");
        return -ENOTTY;
    }
    if(_IOC_NR(cmd) > RS_IOC_MAXNR) {
        PDEBUG("ioctl_error!\n");
        return -ENOTTY;
    }
#ifdef KERN_2_6
/*
    if(_IOC_DIR(cmd) & _IOC_READ) {
        acc = access_ok(VERIFY_WRITE, (void __user *)arg, _IOC_SIZE(cmd));
    } else {
        acc = access_ok(VERIFY_READ, (void __user *)arg, _IOC_SIZE(cmd));
    }

    if(!acc) {
        PDEBUG("ioctl_error!\n");
        return -EFAULT;
    }
*/
#endif
    rs485 = fp->private_data;

    switch(cmd) {
        case RS_RESET:
            if( down_interruptible(&rs485->sem)) {
                up(&rs485->sem);
                return -EINTR;
            }
            err = rs485_reset(rs485);
            udelay(SEM_WAIT);
            up(&rs485->sem);
            return err;

        case RS_FIFO_FLUSH:
            return SUCCESS;

        case RS_SET_BAUDRATE:
            if( down_interruptible(&rs485->sem)) {
                up(&rs485->sem);
                return -EINTR;
            }
            err = rs485_test_device(rs485);
            if(err >= 0 ) {
                rs485->baudrate = (int)arg;
                err = rs485_set_baudrate(rs485);
            }
            udelay(SEM_WAIT);
            up(&rs485->sem);
            return err;

        case RS_GET_BAUDRATE:
            return rs485->baudrate;

        case RS_BLNUMBER_RESET:
            return SUCCESS;

        case RS_GET_VERS:
            rest = copy_to_user((char*)arg, RS485_UTS_RELEASE, sizeof(RS485_UTS_RELEASE));
            if(rest != 0) {
                PDEBUG("rs485_ioctl: get_vers: copy_to_user rest=%ld\n", rest);
                return -EFAULT;
            }
            return RS485_VERSION_CODE;

        case RS_GET_DATAFIFO_SIZE:
            return SIZE_STREAM_FIFO;

        default: break;
    }
    return -ERSNOIOCTL;
}

// ---------------------------------------------------------------------

// --------    --------------------------------
struct file_operations rs485_fops = {
    .owner        = THIS_MODULE,
    .open         = rs485_open,
    .release      = rs485_close,
    .read         = rs485_read,
    .write        = rs485_write,
    .ioctl        = rs485_ioctl,
};

/* --------------------------------------------------------------- */

void rs485_free_mem1(void) {
    kfree(info_port1);
}

void rs485_free_mem2(void) {
    kfree(info_port2);
}

void rs485_delete_card(void) {
    rs485_release_resource();
#ifndef KERN_2_6
    if ( (unregister_chrdev(major, ADAPTER_ID)) !=0){
        RS_PRINT(KERN_CRIT "%s: Error unload module\n", ADAPTER_ID);
    }
#else
    cdev_del(cdev);
    unregister_chrdev_region(dev,rs_nr_dev);
#endif
    return;
}

/* -------------------------------------------------
 *      
 *  :
 * io  <base address>
 * irq <interrupt>
 * --------------------------------------------------- */
 
MODULE_AUTHOR("Andre Zaitsev, Olga Voziakova (ovoz@sai.msu.ru)");
MODULE_DESCRIPTION("rs485 module");
#ifdef MODULE_LICENSE
    MODULE_LICENSE("GPL");
#endif

MODULE_PARM_DESC(io1, "Base COM1-port (hex), default io1=_IO_COM1");
MODULE_PARM_DESC(io2, "Base COM2-port (hex), default io2=_IO_COM2");
MODULE_PARM_DESC(irq1, "Interrupt COM1-port, default irq1=_IRQ_COM1");
MODULE_PARM_DESC(irq2, "Interrupt COM2-port, default irq2=_IRQ_COM2");

MODULE_PARM_DESC(io1, "i");
MODULE_PARM_DESC(io2, "i");
MODULE_PARM_DESC(irq1, "i");
MODULE_PARM_DESC (irq2, "i");

//MODULE_PARM_DESC(portnumb, "PORT NUMBER, default portnumb=_PORT_NUMB");
//MODULE_PARM(portnumb, "i");

MODULE_PARM_DESC(baudrate1, "BAUD RATE UART port1, default baudrate=_BAUDRATE_1");
MODULE_PARM_DESC(baudrate2, "BAUD RATE UART port2, default baudrate=_BAUDRATE_2");
MODULE_PARM_DESC(baudrate1, "i");
MODULE_PARM_DESC(baudrate2, "i");

#ifndef KERN_2_6
EXPORT_NO_SYMBOLS;
#endif

int init_module(void) {
    int err = 0;
//    int err1 = 0;
//    int err2 = 0;
    char version[]=RS485_UTS_RELEASE;

    baudrate1 = baudrate1 ? baudrate1 : _BAUDRATE_1;
    baudrate2 = baudrate2 ? baudrate2 : _BAUDRATE_2;

//    PDEBUG("baudrate1=%d baudrate2=%d\n", baudrate1, baudrate2);

    io1 = io1 ? io1 : rs485_getIO(1);
    io2 = io2 ? io2 : rs485_getIO(2);
    irq1 = irq1 > 0 ? irq1 : rs485_getIRQ(1);
    irq2 = irq2 > 0 ? irq2 : rs485_getIRQ(2);

    info_port1 = (RS485INFO *) kmalloc (sizeof (RS485INFO), GFP_KERNEL);
    if(info_port1 == NULL){
        err = -ENOMEM;
        goto FAIL_1;
    }
    info_port2 = (RS485INFO *) kmalloc (sizeof (RS485INFO), GFP_KERNEL);
    if(info_port2 == NULL){
        err = -ENOMEM;
        goto FAIL_2;
    }
    
    memset((void *)info_port1, 0 , sizeof(RS485INFO));
    memset((void *)info_port2, 0 , sizeof(RS485INFO));


    info_port1->base = io1;
    info_port2->base = io2;

    info_port1->irq = irq1;
    info_port2->irq = irq2;

    info_port1->baudrate = baudrate1;
    info_port2->baudrate = baudrate2;
    
    if( (err=rs485_init()) != 0 ) goto FAIL_3;

    rs485_stop_interrupt_card(info_port1);
    rs485_stop_interrupt_card(info_port2);

    init_MUTEX(&info_port1->sem);        // 2.4 ???
    init_MUTEX(&info_port2->sem);

//    

    RS_PRINT(KERN_INFO "%s module, version %s\n",
                ADAPTER_ID, version);
    RS_PRINT(KERN_INFO "%s: install card on the base=0x%X, 0x%X and irq=%d, %d\n",
                ADAPTER_ID, io1, io2, irq1, irq2);

    return SUCCESS;

//FAIL_5:    //    create_fifo failed
    rs485_delete_card();
FAIL_3:
    rs485_free_mem2();
FAIL_2:
    rs485_free_mem1();
FAIL_1:
    return(err);
}

void cleanup_module(void){
    rs485_free_mem2();
    rs485_free_mem1();
    rs485_delete_card();
    return;
}
